import torch 
from torch.distributions.kl import kl_divergence

from CytoOne.z_w_class import p_pretrain_z_class, p_pretrain_w_class, q_pretrain_z_class, q_pretrain_w_class
from CytoOne.y_class import p_y_class
from CytoOne.base_class import model_base_class

from typing import Optional, Union


class pretrain_model(model_base_class):
    def __init__(self,
                 y_scale: float=0.001,
                 model_device: Optional[Union[str, torch.device]]=None) -> None:
        super().__init__()
        
        if model_device is None:
            self.model_device = torch.device(
                'cuda:0' if torch.cuda.is_available() else 'cpu')
        elif isinstance(model_device, str):
            self.model_device = torch.device(model_device)
        else:
            self.model_device = model_device
        
        self.p_z = p_pretrain_z_class(model_device=self.model_device)
        self.p_w = p_pretrain_w_class(model_device=self.model_device)
        self.p_y = p_y_class(model_device=self.model_device,
                             y_scale=y_scale)
        self.q_z = q_pretrain_z_class(model_device=self.model_device)
        self.q_w = q_pretrain_w_class(model_device=self.model_device)
               
    def _update_distributions(self, 
                              y: torch.tensor):
        q_z_dict = self.q_z(y=y)
        z_samples = self.q_z.get_samples()
        
        q_w_dict = self.q_w(y=y)
        w_samples = self.q_w.get_samples()
        
        p_z_dict = self.p_z()
        p_w_dict = self.p_w()
        p_y_dict = self.p_y(z=z_samples['z'],
                            w=w_samples['w'])
        log_likelihood = p_y_dict['y'].log_prob(y)
        
        return {
            "q_z_dict": q_z_dict,
            "q_w_dict": q_w_dict,
            "p_z_dict": p_z_dict,
            "p_w_dict": p_w_dict,
            "p_y_dict": p_y_dict,
            "log_likelihood": log_likelihood
        } 
    
    def compute_loss(self, 
                     distribution_dict: dict,
                     show_details: bool=False):
        kl_z = kl_divergence(distribution_dict['q_z_dict']['z'], 
                             distribution_dict["p_z_dict"]['z']).mean()
        kl_w = kl_divergence(distribution_dict['q_w_dict']['w'],
                             distribution_dict["p_w_dict"]['w']).mean()
        reconstruction_error = distribution_dict['log_likelihood'].mean()

        elbo = reconstruction_error - kl_w - kl_z
        if show_details:
            print("="*25)
            print("likelihood is {}, kl_z is {}, kl_w is {}".format(reconstruction_error,
                                                                    kl_z,
                                                                    kl_w))
            print("="*25)
        return -elbo
    


# Data IO 
import os 
from copy import deepcopy
import numpy as np 
import pandas as pd 

# PyTorch
import torch  
from torch.distributions.kl import register_kl
from pyro.distributions import Delta 
from torch.distributions import Normal, Categorical
from CytoOne.basic_distributions import zero_inflated_lognormal

# Typing 
from typing import Optional, Tuple, Union


################################
# New KL divergences 
# The formula is KL(p||q) = -\int p \log \dfrac{q}{p} 

@register_kl(Delta, Delta)
def _kl_delta_delta(p, q) -> torch.tensor:
    return -q.log_density 

@register_kl(Delta, zero_inflated_lognormal)
def _kl_delta_ziln(p, q) -> torch.tensor:
    return -q.log_prob(p.v) 

@register_kl(Delta, Categorical)
def _kl_delta_categorical(p, q) -> torch.tensor:
    return -q.log_prob(p.v) 


############################
def data_curation(user_data: Union[pd.DataFrame, str], 
                  batch_index_col_name: Optional[str]=None,
                  condition_index_col_name: Optional[str]=None,
                  subject_index_col_name: Optional[str]=None,
                  cell_type_index_col_name: Optional[str]=None,
                  other_nonprotein_col_names: Optional[list]=None,
                  batches_to_retain: Optional[list]=None,
                  conditions_to_retain: Optional[list]=None,
                  subjects_to_retain: Optional[list]=None,
                  cell_types_to_retain: Optional[list]=None,
                  is_zero_inflated: bool=True,
                  arcsinh5_transform: bool=True,
                  **kwargs) -> Tuple[pd.DataFrame, dict]:
    """Curate CyTOF dataset
    
    This function curates a CyTOF dataset and returns the curated data and 
    the meta information needed for model input 

    Parameters
    ----------
    user_data : Union[pd.DataFrame, str]
        Either a pandas dataframe of or the path to the dataset to be used 
    batch_index_col_name : Optional[str], optional
        The name of the column that indexes batches. 
        If None, we assume all data is from one batch, by default None
    condition_index_col_name : Optional[str], optional
        The name of the column that indexes conditions (treatments). 
        If None, we assume all data is from one condition, by default None
    subject_index_col_name : Optional[str], optional
        The name of the column that indexes subjects (patients). 
        If None, we assume all data is from one subject, by default None
    cell_type_index_col_name : Optional[str], optional
        The name of the column that indexes cell types. 
        This is only used for validation/visualization, by default None
    other_nonprotein_col_names : Optional[list], optional
        The names of columns that provide other information that are not going to be considered by the model. 
        If None, all columns that are not batch, condition, subject, or cell type will be deemed as protein channels, by default None
    batches_to_retain : Optional[list], optional
        A list of batch indices to retain. If None, all batches are kept, by default None
    conditions_to_retain : Optional[list], optional
        A list of condition indices to retain. If None, all conditions are kept, by default None
    subjects_to_retain : Optional[list], optional
        A list o subject indices to retain. If None, all subjects are kept, by default None
    cell_types_to_retain : Optional[list], optional
        A list of cell types to retain. If None, all cell types are kept, by default None
    is_zero_inflated : bool, optional
        Whether or not the data is zero inflated. If True, we will truncate the data from below 0, by default True
    arcsinh5_transform : bool, optional
        Whether or not we need to transform the data using the inverse hyperbolic sine function with factor of 5, by default True

    Returns
    -------
    Tuple[pd.DataFrame, dict]
        The first output is the curated dataframe, while the second one contains meta information of the data. 
    """
    
    # We first make sure that the input is legitmate 
    assert isinstance(batch_index_col_name, str) or (batch_index_col_name is None), "batch_index_col_name should be a string or None"
    assert isinstance(condition_index_col_name, str) or (condition_index_col_name is None), "condition_index_col_name should be a string or None"
    assert isinstance(subject_index_col_name, str) or (subject_index_col_name is None), "subject_index_col_name should be a string or None"
    assert isinstance(cell_type_index_col_name, str) or (cell_type_index_col_name is None), "cell_type_index_col_name should be a string or None"
    assert isinstance(other_nonprotein_col_names, list) or (other_nonprotein_col_names is None), "other_nonprotein_col_names should be a list of strings or None"
    # If ~~~to_retain is specified but the corresponding column names are not. 
    assert (batches_to_retain is None) or (batch_index_col_name is not None), "Please specify batch index col"
    assert (conditions_to_retain is None) or (condition_index_col_name is not None), "Please specify condition index col"
    assert (subjects_to_retain is None) or (subject_index_col_name is not None), "Please specify subject index col"
    assert (cell_types_to_retain is None) or (cell_type_index_col_name is not None), "Please specify cell type index col"
    # Read data 
    if isinstance(user_data, str):
        df = pd.read_csv(user_data, **kwargs)
    else:
        df = deepcopy(user_data)
    # We filter the data frame according to ~~to_retain
    if batches_to_retain is not None:
        df = df[df[batch_index_col_name].isin(batches_to_retain)].reset_index(drop=True)
    if conditions_to_retain is not None:
        df = df[df[condition_index_col_name].isin(conditions_to_retain)].reset_index(drop=True)
    if subjects_to_retain is not None:
        df = df[df[subject_index_col_name].isin(subjects_to_retain)].reset_index(drop=True)
    if cell_types_to_retain is not None:
        df = df[df[cell_type_index_col_name].isin(cell_types_to_retain)].reset_index(drop=True)
    # As various factors might not be labels like 0, 1, 2, ....
    # we convert the indices to nature numbers 
    nonprotein_col_names = []
    if batch_index_col_name is not None:
        batch_index, _ = pd.factorize(df[batch_index_col_name])
        nonprotein_col_names.append(batch_index_col_name)
    else:
        batch_index = np.zeros(df.shape[0], dtype=np.int32)
    if condition_index_col_name is not None:
        condition_index, _ = pd.factorize(df[condition_index_col_name])
        nonprotein_col_names.append(condition_index_col_name)
    else:
        condition_index = np.zeros(df.shape[0], dtype=np.int32)
    if subject_index_col_name is not None:
        subject_index, _ = pd.factorize(df[subject_index_col_name])
        nonprotein_col_names.append(subject_index_col_name)
    else:
        subject_index = np.zeros(df.shape[0], dtype=np.int32)
    if cell_type_index_col_name is not None:
        cell_type_index, _ = pd.factorize(df[cell_type_index_col_name])
        nonprotein_col_names.append(cell_type_index_col_name)
    else:
        cell_type_index = np.zeros(df.shape[0], dtype=np.int32)
    
    if other_nonprotein_col_names is not None:
        nonprotein_col_names += other_nonprotein_col_names
    # Everything else is then protein channels
    protein_col_names = list(df.columns[~df.columns.isin(nonprotein_col_names)])
    # Then we drop all nonprotein columns 
    df.drop(columns=nonprotein_col_names, inplace=True)
    # Transform data 
    if arcsinh5_transform:
        df = np.arcsinh(df/5)
    # Make sure that the data is zero inflated 
    if is_zero_inflated:
        df = np.clip(df, a_min=0, a_max=None)
    # Then we append the curated indices to the dataframe 
    df['batch_index'] = batch_index
    df['condition_index'] = condition_index
    df['subject_index'] = subject_index
    df['cell_type_index'] = cell_type_index
    # Calculate some meta information 
    n_b = df['batch_index'].max(axis=0)+1
    n_c = df['condition_index'].max(axis=0)+1
    n_s = df['subject_index'].max(axis=0)+1
    n_types = df['cell_type_index'].max(axis=0)+1
    
    meta_dict = {
        "batch_index_col_name": batch_index_col_name,
        "condition_index_col_name": condition_index_col_name,
        "subject_index_col_name": subject_index_col_name,
        "cell_type_index_col_name": cell_type_index_col_name,
        "other_nonprotein_col_names": other_nonprotein_col_names,
        "protein_col_names": protein_col_names,
        "n_batches": n_b,
        "n_conditions": n_c,
        "n_subjects": n_s,
        "n_types": n_types,
        "N": df.shape[0],
        "y_dim": len(protein_col_names)
    }
    
    return df, meta_dict
    
    
def generate_cytoone_model_input(meta_dict: dict,
                                 x_dim: int, 
                                 y_scale: float,
                                 n_cell_types: int,
                                 model_check_point_path: Optional[str]=None,
                                 model_device: Optional[Union[str, torch.device]]=None) -> dict:
    """Generate model input for CytoOne 

    By using the meta information we extracted from the data, the functions 
    constructs a dictionary as model input 
    
    Parameters
    ----------
    meta_dict : dict
        A dictionary containing meta information of the data
    x_dim : int
        Dimensions of the latent space
    y_scale: float
        The scale for the noise added to y
    n_cell_types : int
        Number of cell clusters 
    model_check_point_path : Optional[str], optional
        The path to the pretrained model. If none, random weights are used, by default None
    model_device: Optional[Union[str, torch.device]]
        The device to use 
    Returns
    -------
    dict
        A dictionary of the model input 
    """
    
    if model_device is None:
        model_device = torch.device(
            'cuda:0' if torch.cuda.is_available() else 'cpu')
    elif isinstance(model_device, str):
        model_device = torch.device(model_device)
    else:
        model_device = model_device
    
    parameter_dict = {
        "y_dim": meta_dict['y_dim'],
        "x_dim": x_dim,
        "y_scale": y_scale,
        "n_batches": meta_dict['n_batches'],
        "n_conditions": meta_dict['n_conditions'],
        "n_subjects": meta_dict['n_subjects'],
        "n_cell_types": n_cell_types,
        "model_check_point_path": model_check_point_path,
        "model_device": model_device
    }
    
    return parameter_dict


def generate_pretrain_data(y_scale: float=0.001,
                           low: float=-10,
                           high: float=2.5,
                           n_sample: int=100000) -> pd.DataFrame:
    """Generate data to be used for pretraining 

    Parameters
    ----------
    y_scale : float, optional
        The scale of the normal distribution, by default 0.001
    low: float
        The lower bound 
    high: float
        The upper bound 
    n_sample : int, optional
        Number of samples, by default 100000

    Returns
    -------
    pd.DataFrame
        A dataframe containing generated pretraining data
    """
    # We first generate data frome a pretty large square 
    z = np.random.uniform(low=low, high=high, size=n_sample)
    w = np.random.uniform(low=low, high=high, size=n_sample)
    # Then we generate truly zero inflated observations 
    pure_y = np.round(1/(1+np.exp(-w))) * np.exp(z)
    # Finally, we corrupt them with a little noise 
    y = pure_y + np.random.normal(loc=0, scale=y_scale, size=n_sample)
    df = pd.DataFrame({"y": y,
                       "z": z, 
                       "w": w})
    return df


# PyTorch 
import torch 
import torch.nn as nn 
import torch.nn.functional as F
from torch.distributions import Normal, Independent
from pyro.distributions import Delta 

from CytoOne.base_class import component_base_class


class p_effect_pi_coef_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 n_conditions: int,
                 n_subjects: int,
                 n_cell_types: int): 
        super().__init__(stage_to_change="abundance effect estimation",
                         distribution_info={"beta": (n_conditions, n_cell_types),
                                            "gamma": (n_subjects, n_cell_types)})
        self.model_device = model_device
        
        self.n_conditions = n_conditions
        self.n_subjects = n_subjects
        self.n_cell_types = n_cell_types
    
    def _update_distributions(self,
                              eta_sample: torch.tensor) -> None:
        gamma_scale = torch.sqrt(torch.exp(eta_sample))
        for dist in self.distribution_dict:
            if (self.distribution_info_dict[dist][0] == 1):
                self.distribution_dict[dist] = Independent(Delta(v=torch.zeros(self.distribution_info_dict[dist],
                                                                               device=self.model_device),
                                                                 log_density=torch.ones(self.distribution_info_dict[dist],
                                                                                        device=self.model_device)),
                                                           reinterpreted_batch_ndims=2) 
            else:
                if dist == "gamma":
                    self.distribution_dict[dist] = Independent(Normal(loc=torch.zeros(self.distribution_info_dict[dist],
                                                                                      device=self.model_device),
                                                                    scale=torch.ones(self.distribution_info_dict[dist],
                                                                                     device=self.model_device)*gamma_scale),
                                                            reinterpreted_batch_ndims=2) 
                else:
                    self.distribution_dict[dist] = Independent(Normal(loc=torch.zeros(self.distribution_info_dict[dist],
                                                                                      device=self.model_device),
                                                                    scale=torch.ones(self.distribution_info_dict[dist],
                                                                                     device=self.model_device)),
                                                            reinterpreted_batch_ndims=2)
    

class q_effect_pi_coef_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 n_conditions: int,
                 n_subjects: int,
                 n_cell_types: int):
        super().__init__(stage_to_change="abundance effect estimation",
                         distribution_info={"beta": (n_conditions, n_cell_types),
                                            "gamma": (n_subjects, n_cell_types)})
        self.model_device = model_device
        self.n_conditions = n_conditions
        self.n_subjects = n_subjects
        self.n_cell_types = n_cell_types
        
        self.parameter_dict = nn.ParameterDict({})

        for dist in self.distribution_dict:
            if (self.distribution_info_dict[dist][0] > 1):
                self.parameter_dict.update({dist+"_loc": nn.Parameter(torch.randn(self.distribution_info_dict[dist]),
                                                                            requires_grad=True),
                                            dist+"_scale": nn.Parameter(torch.randn(self.distribution_info_dict[dist]),
                                                                            requires_grad=True)})
        
    def _update_distributions(self):
        for dist in self.distribution_dict:
            if (self.distribution_info_dict[dist][0] == 1):
                self.distribution_dict[dist] = Independent(Delta(v=torch.zeros(self.distribution_info_dict[dist],
                                                                               device=self.model_device),
                                                                 log_density=torch.ones(self.distribution_info_dict[dist],
                                                                                        device=self.model_device)),
                                                           reinterpreted_batch_ndims=2)
            else:
                self.distribution_dict[dist] = Independent(Normal(loc=self.parameter_dict[dist+'_loc'],
                                                                  scale=F.softplus(self.parameter_dict[dist+'_scale'], beta=1)+0.00001),
                                                            reinterpreted_batch_ndims=2)


class p_effect_pi_hyper_coef_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 n_subjects: int,
                 n_cell_types: int):
        super().__init__(stage_to_change="abundance effect estimation",
                         distribution_info={"eta": (1, n_cell_types)})
        self.model_device = model_device
        self.n_subjects = n_subjects
        self.n_cell_types = n_cell_types
        
    def _update_distributions(self):
        for dist in self.distribution_dict:
            self.distribution_dict[dist] = Independent(Normal(loc=torch.zeros((1, self.n_cell_types),
                                                                              device=self.model_device),
                                                              scale=torch.ones((1, self.n_cell_types),
                                                                               device=self.model_device)),
                                                       reinterpreted_batch_ndims=2)  
        
        
class q_effect_pi_hyper_coef_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 n_subjects: int,
                 n_cell_types: int):
        super().__init__(stage_to_change="abundance effect estimation",
                         distribution_info={"eta": (1, n_cell_types)})       
        self.model_device = model_device
        self.n_subjects = n_subjects
        self.n_cell_types = n_cell_types
        
        self.parameter_dict = nn.ParameterDict({})
        if self.n_subjects > 1:
            for dist in self.distribution_dict:
                self.parameter_dict.update({dist+"_loc": nn.Parameter(torch.randn(self.distribution_info_dict[dist]),
                                                                            requires_grad=True),
                                            dist+"_scale": nn.Parameter(torch.randn(self.distribution_info_dict[dist]),
                                                                            requires_grad=True)})
    
    def _update_distributions(self):
        if self.n_subjects > 1:
            for dist in self.distribution_dict:
                self.distribution_dict[dist] = Independent(Normal(loc=self.parameter_dict[dist+'_loc'],
                                                                  scale=F.softplus(self.parameter_dict[dist+'_scale'], beta=1)+0.00001),
                                                        reinterpreted_batch_ndims=2)
        else:
            for dist in self.distribution_dict:
                self.distribution_dict[dist] = Independent(Normal(loc=torch.zeros(self.distribution_info_dict[dist],
                                                                                  device=self.model_device),
                                                                scale=torch.ones(self.distribution_info_dict[dist],
                                                                                 device=self.model_device)),
                                                        reinterpreted_batch_ndims=2)  



import torch 
import torch.nn as nn 
import torch.nn.functional as F
from torch.distributions import Normal, Independent

from CytoOne.base_class import component_base_class

import numpy as np 
from collections import OrderedDict

from typing import Union, Optional


class p_x_class(component_base_class):
    def __init__(self, 
                 model_device: torch.device,
                 x_dim: int) -> None:
        self.model_device = model_device
        self.x_dim = x_dim
        super().__init__(stage_to_change="dimension reduction",
                         distribution_info={"x": None})
        
    def _update_distributions(self,
                              embedding: Optional[torch.tensor]=None):
        if embedding is None:
            self.distribution_dict["x"] = Independent(Normal(loc=torch.zeros(self.x_dim,
                                                                             device=self.model_device),
                                                             scale=torch.ones(self.x_dim,
                                                                             device=self.model_device)), 
                                                        reinterpreted_batch_ndims=1) 
        else: 
            self.distribution_dict["x"] = Independent(Normal(loc=embedding,
                                                              scale=torch.ones(self.x_dim,
                                                                               device=self.model_device)), 
                                                        reinterpreted_batch_ndims=1) 
        
    
        
class q_x_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 y_dim: int,
                 x_dim: int,
                 n_batches: int=1,
                 n_conditions: int=1,
                 n_subjects: int=1) -> None:
        self.model_device = model_device
        self.x_dim = x_dim
        super().__init__(stage_to_change="dimension reduction",
                         distribution_info={"x": None})
            
        extra_n_dim = np.sum([n for n in [n_batches, n_conditions, n_subjects] if n>1], dtype=int)
        
        self.loc_mapping = nn.Sequential(OrderedDict([
            ('fc1', nn.Linear(in_features=y_dim + extra_n_dim,
                              out_features=512,
                              bias=True)),
            ('fc1_bn', nn.BatchNorm1d(512)),
            ('fc1_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc2', nn.Linear(in_features=512,
                              out_features=256,
                              bias=True)),
            ('fc2_bn', nn.BatchNorm1d(256)),
            ('fc2_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc3', nn.Linear(in_features=256,
                              out_features=128,
                              bias=True)),
            ('fc3_bn', nn.BatchNorm1d(128)),
            ('fc3_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc4', nn.Linear(in_features=128,
                              out_features=x_dim,
                              bias=True))
        ]))
        self.log_scale_mapping = nn.Sequential(OrderedDict([
            ('fc1', nn.Linear(in_features=y_dim + extra_n_dim,
                              out_features=512,
                              bias=True)),
            ('fc1_bn', nn.BatchNorm1d(512)),
            ('fc1_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc2', nn.Linear(in_features=512,
                              out_features=256,
                              bias=True)),
            ('fc2_bn', nn.BatchNorm1d(256)),
            ('fc2_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc3', nn.Linear(in_features=256,
                              out_features=128,
                              bias=True)),
            ('fc3_bn', nn.BatchNorm1d(128)),
            ('fc3_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc4', nn.Linear(in_features=128,
                              out_features=x_dim,
                              bias=True))
        ]))
           
    def _update_distributions(self, 
                              z: torch.tensor, 
                              w: torch.tensor,
                              FB: torch.tensor,
                              FC: torch.tensor,
                              RS: torch.tensor):
        # z_w = (z+w)/2
        w_temp = torch.round(F.sigmoid(w))
        wq = w + (w_temp - w).detach()
        
        z = torch.exp(z)
        z_w = wq * z 
        
        effect_list = [m for m in [FB, FC, RS] if m.shape[1] > 1]
        
        z_w_with_effct = torch.cat([z_w] + effect_list, dim=1)
        
        self.distribution_dict['x'] = Independent(Normal(loc=self.loc_mapping(z_w_with_effct),
                                                         scale=F.softplus(self.log_scale_mapping(z_w_with_effct)) + 0.00001),
                                                         reinterpreted_batch_ndims=1)
        

    


"""The Command-Line Interface (CLI) of CytoOne

The CLI of CytoOne can be accessed via ``python -m CytoOne``.

:Example:

    Get help:
    
    .. code-block:: bash

        python -m CytoOne -h
    
    Check version and authors:
    
    .. code-block:: bash
    
        python -m CytoOne --version 
        python -m CytoOne --author

**Pros**
    * Shallow learning curve
    
**Cons**
    * Fewer configurations 
    * No inspections of intermediate results
"""


import os
import sys
import torch
import argparse

from CytoOne import __version__, __author__

parser = argparse.ArgumentParser(description="CytoOne")

parser.add_argument("--version", action="version",
                    version=__version__, help="Display the version of the software")
parser.add_argument("--author", action="version", version=__author__,
                    help="Check the author list of the algorithm")


# PyTorch 
import torch 
import torch.nn.functional as F
from torch.distributions import Normal, Independent

# Base class 
from CytoOne.base_class import component_base_class


class p_y_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 y_dim: int=1,
                 y_scale: float=0.001) -> None:
        """p(y|z, w)

        This component defines how the observations are generated 
        given the latent expressions and zero-inflations 
        
        Parameters
        ----------
        y_dim : int, optional
            The dimension of y. During pretrain stage, we are concerned with 
            reconstructing y. Therefore, we "flatten" ys and thus the dimension is 1.
            However, during training, the dimension should match with the number of 
            protein markers in the dataset, by default 1
        y_scale : float, optional
            This defines how much noise should be used to corrupt the generated 
            observations. This is chosen to prevent -\infty log-likelihood from 
            being generated due to negative values, by default 0.001
        """
        super().__init__(stage_to_change="pretrain",
                         distribution_info={"y": None})
        self.model_device = model_device
        self.y_dim = y_dim
        self.y_scale = torch.tensor(y_scale, 
                                    dtype=torch.float32,
                                    device=self.model_device)
        self.reinterpreted_batch_ndims = 0
        if self.y_dim > 1:
            self.reinterpreted_batch_ndims = 1
    
    def _update_distributions(self,
                              z: torch.tensor,
                              w: torch.tensor) -> None:
        """Update p(y|z, w)

        Parameters
        ----------
        z : torch.tensor
            The latent expressions 
        w : torch.tensor
            The latent zero-inflations 
        """
        # We first transform w to 0 and 1
        w_temp = torch.round(F.sigmoid(w))
        # Like VQ-VAE, we use straight-through estimator
        wq = w + (w_temp - w).detach()
        
        # This generates a log-normal RV 
        z = torch.exp(z)
        # Multiplying two yields a zero-inflated log-normal
        denoised_y = wq * z 
        
        # The actual observations will then be corrupted by a little 
        # Gaussian noise 
        # NOTE: in the future, we will future consider other types of 
        # noise distributions and explore the possibility of inferring 
        # noise level via SGD. For now, we will fix the normal formulation.
        self.distribution_dict['y'] = Independent(Normal(loc=denoised_y,
                                                         scale=self.y_scale),
                                                  reinterpreted_batch_ndims=self.reinterpreted_batch_ndims)

        


# Numeric manipulation 
import math
import numpy as np 

# PyTorch
import torch
import torch.nn as nn 
import torch.nn.functional as F
import torch.distributions as d 
from torch.distributions import constraints
from torch.distributions.utils import (
    broadcast_all,
    lazy_property,
    logits_to_probs,
    probs_to_logits,
    _standard_normal,
)
from torch.nn.functional import softplus

# Pyro 
from pyro.distributions import TorchDistribution
from pyro.distributions.util import broadcast_shape
from pyro.distributions import Gamma as pyro_Gamma
from pyro.distributions import LogNormal as pyro_LogNormal


class mollified_uniform(TorchDistribution):
    arg_constraints = {'uniform_low': constraints.real,
                       'uniform_up': constraints.real,
                       'normal_scale': constraints.positive}
    support = constraints.real
    has_rsample = True

    def __init__(self, uniform_low, uniform_up, normal_scale, *, validate_args=None):
        self.uniform_low, self.uniform_up, self.normal_scale = broadcast_all(uniform_low, uniform_up, normal_scale)
        super().__init__(self.uniform_low.shape, validate_args=validate_args)

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(mollified_uniform, _instance)
        batch_shape = torch.Size(batch_shape)
        new.uniform_low = self.uniform_low.expand(batch_shape)
        new.uniform_up = self.uniform_up.expand(batch_shape)
        new.normal_scale = self.normal_scale.expand(batch_shape)
        super(mollified_uniform, new).__init__(batch_shape, validate_args=False)
        new._validate_args = self._validate_args
        return new 
    
    def log_prob(self, value, offset=0.01):
        if self._validate_args:
            self._validate_sample(value) 
        denum = self.uniform_up - self.uniform_low
        upper_cdf = 0.5 * (1 + torch.erf((self.uniform_up - value) * self.normal_scale.reciprocal() / math.sqrt(2)))
        lower_cdf = 0.5 * (1 + torch.erf((self.uniform_low - value) * self.normal_scale.reciprocal() / math.sqrt(2)))
        return ((upper_cdf - lower_cdf) * denum.reciprocal() + offset).log()
        #return torch.asinh((upper_cdf - lower_cdf) * denum.reciprocal()) - 1

    def rsample(self, sample_shape=torch.Size()):
        shape = self._extended_shape(sample_shape)
        # Sample from standard uniform first 
        rand = torch.rand(shape, dtype=self.uniform_low.dtype, device=self.uniform_low.device)
        u = rand * (self.uniform_up - self.uniform_low) + self.uniform_low
        eps = _standard_normal(shape, dtype=self.normal_scale.dtype, device=self.normal_scale.device)
        return u + eps * self.normal_scale
    
    @property
    def mean(self):
        return (self.uniform_up + self.uniform_low)/2

    @property
    def variance(self):
        temp = self.uniform_up + self.uniform_low
        return temp.pow(2)/12 + self.normal_scale.pow(2)


class zero_inflated_positive_distribution(TorchDistribution):
    """
    Generic Zero Inflated positive distribution.

    This can be used directly or can be used as a base class as e.g. for
    :class:`ZeroInflatedPoisson` and :class:`ZeroInflatedNegativeBinomial`.

    :param TorchDistribution base_dist: the base distribution.
    :param torch.Tensor gate: probability of extra zeros given via a Bernoulli distribution.
    :param torch.Tensor gate_logits: logits of extra zeros given via a Bernoulli distribution.
    """

    arg_constraints = {
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
    }

    def __init__(self, base_dist, *, gate=None, gate_logits=None, validate_args=None):
        if (gate is None) == (gate_logits is None):
            raise ValueError(
                "Either `gate` or `gate_logits` must be specified, but not both."
            )
        if gate is not None:
            batch_shape = broadcast_shape(gate.shape, base_dist.batch_shape)
            self.gate = gate.expand(batch_shape)
        else:
            batch_shape = broadcast_shape(gate_logits.shape, base_dist.batch_shape)
            self.gate_logits = gate_logits.expand(batch_shape)
        if base_dist.event_shape:
            raise ValueError(
                "ZeroInflatedDistribution expected empty "
                "base_dist.event_shape but got {}".format(base_dist.event_shape)
            )

        self.base_dist = base_dist.expand(batch_shape)
        event_shape = torch.Size()

        super().__init__(batch_shape, event_shape, validate_args)

    @constraints.dependent_property
    def support(self):
        return self.base_dist.support

    @lazy_property
    def gate(self):
        return logits_to_probs(self.gate_logits)

    @lazy_property
    def gate_logits(self):
        return probs_to_logits(self.gate)

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)

        if "gate" in self.__dict__:
            gate, value = broadcast_all(self.gate, value)
            log_prob = self.base_dist.log_prob(value+0.00001)
            log_prob = torch.where(value == 0, (gate).log(), (-gate).log1p() + log_prob)
        else:
            gate_logits, value = broadcast_all(self.gate_logits, value)
            log_prob_minus_log_gate = -gate_logits + self.base_dist.log_prob(value)
            log_gate = -softplus(-gate_logits)
            log_prob = log_prob_minus_log_gate + log_gate
            zero_log_prob = softplus(log_prob_minus_log_gate) + log_gate
            log_prob = torch.where(value == 0, zero_log_prob, log_prob)
        return log_prob

    def sample(self, sample_shape=torch.Size()):
        shape = self._extended_shape(sample_shape)
        with torch.no_grad():
            mask = torch.bernoulli(self.gate.expand(shape)).bool()
            samples = self.base_dist.expand(shape).sample()
            samples = torch.where(mask, samples.new_zeros(()), samples)
        return samples


    @lazy_property
    def mean(self):
        return (1 - self.gate) * self.base_dist.mean

    @lazy_property
    def variance(self):
        return (1 - self.gate) * (
            self.base_dist.mean**2 + self.base_dist.variance
        ) - (self.mean) ** 2

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(type(self), _instance)
        batch_shape = torch.Size(batch_shape)
        gate = self.gate.expand(batch_shape) if "gate" in self.__dict__ else None
        gate_logits = (
            self.gate_logits.expand(batch_shape)
            if "gate_logits" in self.__dict__
            else None
        )
        base_dist = self.base_dist.expand(batch_shape)
        zero_inflated_positive_distribution.__init__(
            new, base_dist, gate=gate, gate_logits=gate_logits, validate_args=False
        )
        new._validate_args = self._validate_args
        return new


class zero_inflated_gamma(zero_inflated_positive_distribution):
    arg_constraints = {
        #"concentration": constraints.positive,
        #"rate": constraints.positive,
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
    }
    support = constraints.greater_than_eq(0)

    def __init__(self, concentration, rate, *, 
                 gate=None, gate_logits=None, validate_args=None):
        base_dist = pyro_Gamma(concentration=concentration, 
                          rate=rate, validate_args=False)
        base_dist._validate_args = validate_args

        super().__init__(
            base_dist, gate=gate, gate_logits=gate_logits, validate_args=validate_args
        )


class zero_inflated_lognormal(zero_inflated_positive_distribution):
    arg_constraints = {
        #"concentration": constraints.positive,
        #"rate": constraints.positive,
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
    }
    support = constraints.greater_than_eq(0)

    def __init__(self, loc, scale, *, 
                 gate=None, gate_logits=None, validate_args=None):
        base_dist = pyro_LogNormal(loc=loc, 
                          scale=scale, validate_args=False)
        base_dist._validate_args = validate_args

        super().__init__(
            base_dist, gate=gate, gate_logits=gate_logits, validate_args=validate_args
        )




import torch 
import torch.nn as nn 
import torch.nn.functional as F
from CytoOne.utilities import _kl_delta_delta, _kl_delta_ziln, _kl_delta_categorical            
from torch.distributions.kl import kl_divergence

from CytoOne.y_class import p_y_class
from CytoOne.z_w_class import p_z_w_class, q_z_w_class
from CytoOne.x_class import p_x_class, q_x_class
from CytoOne.pi_class import p_pi_class, q_pi_class

from CytoOne.base_class import model_base_class

from typing import Optional, Union


class cytoone_model(model_base_class):
    def __init__(self,
                 y_dim: int,
                 x_dim: int, 
                 y_scale: float,
                 n_batches: int,
                 n_conditions: int,
                 n_subjects: int,
                 n_cell_types: int,
                 model_check_point_path: Optional[str]=None,
                 model_device: Optional[Union[str, torch.device]]=None) -> None:
        super().__init__()
        
        if model_device is None:
            self.model_device = torch.device(
                'cuda:0' if torch.cuda.is_available() else 'cpu')
        elif isinstance(model_device, str):
            self.model_device = torch.device(model_device)
        else:
            self.model_device = model_device
        
        self.n_batches = n_batches
        self.n_conditions = n_conditions
        self.n_subjects = n_subjects
        self.n_cell_types = n_cell_types
        self.x_dim = x_dim
        self.current_stage = None
        
        self.p_y = p_y_class(model_device=self.model_device,
                             y_dim=y_dim,
                             y_scale=y_scale)
        self.p_z_w = p_z_w_class(model_device=self.model_device,
                                 y_dim=y_dim,
                                 x_dim=x_dim,
                                 n_batches=n_batches,
                                 n_conditions=n_conditions,
                                 n_subjects=n_subjects)
        self.q_z_w = q_z_w_class(model_device=self.model_device,
                                 y_dim=y_dim)
        if model_check_point_path is not None:
            self.q_z_w.load_pretrained_model(model_check_point_path=model_check_point_path)
            
        self.p_x = p_x_class(model_device=self.model_device,
                             x_dim=x_dim)
        self.q_x = q_x_class(model_device=self.model_device,
                             y_dim=y_dim,
                             x_dim=x_dim,
                             n_batches=n_batches,
                             n_conditions=n_conditions,
                             n_subjects=n_subjects)
                
        self.p_pi = p_pi_class(model_device=self.model_device,
                               n_cell_types=n_cell_types,
                               n_conditions=n_conditions,
                               n_subjects=n_subjects)
        self.q_pi = q_pi_class(model_device=self.model_device,
                               x_dim=x_dim,
                               n_cell_types=n_cell_types,
                               vq_vae_weight=0.25)
        
    def initialize_pi_distributions(self,
                                    n_cell_types: int):
        self.n_cell_types = n_cell_types
        self.p_pi = p_pi_class(model_device=self.model_device,
                               n_cell_types=n_cell_types,
                               n_conditions=self.n_conditions,
                               n_subjects=self.n_subjects)
        self.q_pi = q_pi_class(model_device=self.model_device,
                               x_dim=self.x_dim,
                               n_cell_types=n_cell_types,
                               vq_vae_weight=0.25)

    def get_posterior_samples(self,
                              y: torch.tensor,
                              FB: torch.tensor,
                              FC: torch.tensor,
                              RS: torch.tensor,
                              mode: str="validation",
                              get_mean: bool=False):
        assert mode in ["training", "validation"], "mode has to be either training or validation"
        if mode == "training":
            self.train()
        else:
            self.eval()
        with torch.set_grad_enabled(mode=="training"):
            # To generate all the distributions, we start from
            # the posteriors 
            # The steps are almost always: generate the distributions
            # Then sample from the distributions 
            ####################################
            # Q 
            ##############
            # z and w
            q_z_w_dict = self.q_z_w(y=y)
            z_w_samples = self.q_z_w.get_samples(get_mean=get_mean)
            # x 
            q_x_dict = self.q_x(z=z_w_samples['z'],
                                w=z_w_samples['w'],
                                FB=FB,
                                FC=FC,
                                RS=RS)
            x_samples = self.q_x.get_samples(get_mean=get_mean)

            q_pi_dict = {}
            if self.current_stage == "clustering":
                q_pi_dict = self.q_pi(x=x_samples['x']) 
        return q_z_w_dict, q_x_dict, q_pi_dict, z_w_samples, x_samples
    
    def get_latent_samples(self,
                           FC: torch.tensor,
                           RS: torch.tensor,
                           q_pi_dict: Optional[dict]=None,
                           mode: str="validation",
                           get_mean: bool=False):
        assert mode in ["training", "validation"], "mode has to be either training or validation"
        if mode == "training":
            self.train()
        else:
            self.eval()
        with torch.set_grad_enabled(mode=="training"):
            # x
            p_pi_dict = {}
            p_pi_samples = None
            p_x_samples = None
            if self.current_stage == "dimension reduction":
                p_x_dict = self.p_x()
                if q_pi_dict is None:
                    p_x_samples = self.p_x.get_samples(get_mean=get_mean)
            elif self.current_stage == "clustering":
                p_pi_dict = self.p_pi(FC=FC,
                                      RS=RS)
                if q_pi_dict is None:
                    p_pi_samples = self.p_pi.get_samples()
                    one_hot_encoding = torch.zeros(FC.size(0), 
                                                   self.q_pi.cell_embeddings.weight.size(0))
                    one_hot_encoding.scatter_(1, p_pi_samples['pi'], 1)
                    
                    embedding = torch.matmul(one_hot_encoding, self.q_pi.cell_embeddings.weight)
                    p_x_dict = self.p_x(embedding=embedding)
                    p_x_samples = self.p_x.get_samples(get_mean=get_mean)
                else:
                    p_x_dict = self.p_x(embedding=q_pi_dict['embedding'])
        
        return p_pi_dict, p_x_dict, p_pi_samples, p_x_samples
    
    def reconstruct_y(self,
                      FB: torch.tensor,
                      FC: torch.tensor,
                      RS: torch.tensor,
                      x_samples: dict, 
                      z_w_samples: Optional[dict]=None, 
                      y: Optional[torch.tensor]=None,
                      mode: str="validation",
                      get_mean: bool=False):
        assert mode in ["training", "validation"], "mode has to be either training or validation"
        if mode == "training":
            self.train()
        else:
            self.eval()
        with torch.set_grad_enabled(mode=="training"):
            # z and w
            p_z_w_dict = self.p_z_w(x=x_samples['x'],
                                    FB=FB,
                                    FC=FC,
                                    RS=RS)
            log_likelihood = 0
            if z_w_samples is None:
                p_z_w_samples = self.p_z_w.get_samples(get_mean=get_mean)
                # y
                p_y_dict = self.p_y(z=p_z_w_samples['z'],
                                    w=p_z_w_samples['w']) 
            else:
                p_z_w_samples = None
                # y
                p_y_dict = self.p_y(z=z_w_samples['z'],
                                    w=z_w_samples['w']) 
                if y is not None:
                    log_likelihood = p_y_dict['y'].log_prob(y)
                
            y_samples = self.p_y.get_samples(get_mean=get_mean)
        return p_z_w_dict, p_y_dict, p_z_w_samples, y_samples, log_likelihood
    
    def _update_distributions(self, 
                              y: torch.tensor,
                              FB: torch.tensor,
                              FC: torch.tensor,
                              RS: torch.tensor):
        q_z_w_dict, q_x_dict, q_pi_dict, z_w_samples, x_samples = self.get_posterior_samples(y=y,
                                                                                             FB=FB,
                                                                                             FC=FC,
                                                                                             RS=RS,
                                                                                             mode="training",
                                                                                             get_mean=False)
        p_pi_dict, p_x_dict, _, _ = self.get_latent_samples(FC=FC,
                                                            RS=RS,
                                                            q_pi_dict=q_pi_dict,
                                                            mode="training",
                                                            get_mean=False)
        p_z_w_dict, p_y_dict, _, _, log_likelihood = self.reconstruct_y(FB=FB,
                                                                        FC=FC,
                                                                        RS=RS,
                                                                        x_samples=x_samples,
                                                                        z_w_samples=z_w_samples,
                                                                        y=y,
                                                                        mode='training',
                                                                        get_mean=False)
            
        return {
            "q_pi_dict": q_pi_dict,
            "p_pi_dict": p_pi_dict,
            "q_x_dict": q_x_dict,
            "p_x_dict": p_x_dict,
            "q_z_w_dict": q_z_w_dict,
            "p_z_w_dict": p_z_w_dict,
            "p_y_dict": p_y_dict,
            "log_likelihood": log_likelihood
        }
    
    def normalize_samples(self,
                          y: torch.tensor,
                          FB: torch.tensor,
                          FC: torch.tensor,
                          RS: torch.tensor,
                          normalize_to_batch: Optional[int]=0,
                          normalize_to_condition: Optional[int]=0,
                          normalize_to_subject: Optional[int]=0,
                          get_mean: bool=False):
        if normalize_to_batch is None:
            nFB = FB
        else:
            nFB = torch.zeros_like(FB)
            nFB[:, normalize_to_batch] = 1
        if normalize_to_condition is None:
            nFC = FC
        else:
            nFC = torch.zeros_like(FC)
            nFC[:, normalize_to_condition] = 1
        if normalize_to_subject is None:
            nRS = RS
        else:
            nRS = torch.zeros_like(RS)
            nRS[:, normalize_to_subject] = 1
        
        self.eval()
        with torch.no_grad():
            _, _, _, _, x_samples = self.get_posterior_samples(y=y,
                                                               FB=FB,
                                                               FC=FC,
                                                               RS=RS,
                                                               mode="validation",
                                                               get_mean=get_mean)
            _, _, _, y_samples, _ = self.reconstruct_y(FB=nFB,
                                                       FC=nFC,
                                                       RS=nRS,
                                                       x_samples=x_samples,
                                                       z_w_samples=None,
                                                       y=None,
                                                       mode='validation',
                                                       get_mean=get_mean)
        return y_samples
            
    def compute_loss(self, 
                     distribution_dict: dict,
                     show_details: bool=False):
        reconstruction_error = distribution_dict['log_likelihood'].mean()
        
        kl_x = kl_divergence(distribution_dict['q_x_dict']['x'],
                            distribution_dict['p_x_dict']['x']).mean()  
        kl_z = kl_divergence(distribution_dict['q_z_w_dict']['z'],
                            distribution_dict['p_z_w_dict']['z']).mean()
        kl_w = kl_divergence(distribution_dict['q_z_w_dict']['w'],
                            distribution_dict['p_z_w_dict']['w']).mean()
        kl_pi = 0
        vq_loss = 0
        if self.current_stage == "clustering":
            kl_pi = kl_divergence(distribution_dict['q_pi_dict']['pi'],
                                  distribution_dict['p_pi_dict']['pi']).mean()
            vq_loss = distribution_dict['q_pi_dict']['vq_loss']
        local_kl = - kl_x - kl_z - kl_w - kl_pi

        elbo = reconstruction_error + local_kl
        if show_details:
            print("="*25)
            print("likelihood is {}, kl_x is {}, kl_z is {}, kl_w is {}, kl_pi is {} vq_loss is {}".format(reconstruction_error,
                                                                                                           kl_x,
                                                                                                           kl_z,
                                                                                                           kl_w,
                                                                                                           kl_pi,
                                                                                                           vq_loss))
            print("="*25)
        return -elbo + vq_loss


# PyTorch 
import torch 
from torch.distributions.kl import kl_divergence

from CytoOne.z_w_class import p_effect_z_w_class
from CytoOne.z_w_effect_coef_class import p_effect_z_w_coef_class, p_effect_z_w_hyper_coef_class,\
                                         q_effect_z_w_coef_class, q_effect_z_w_hyper_coef_class

from CytoOne.base_class import model_base_class

from typing import Optional, Union

class expression_model(model_base_class):
    def __init__(self,
                 y_dim: int,
                 n_batches: int,
                 n_conditions: int,
                 n_subjects: int,
                 n_cell_types: int,
                 effect_type: str,
                 model_device: Optional[Union[str, torch.device]]=None):
        super().__init__()
        
        assert (n_cell_types > 1) or (n_batches > 1) or (n_conditions > 1) or (n_subjects > 1), "You need to have at least 2 cell types or 2 batches or 2 conditions or 2 subjects to estimate some effects."
        assert effect_type in ["expression", "inflation"], "Illegal type..."
        
        if model_device is None:
            self.model_device = torch.device(
                'cuda:0' if torch.cuda.is_available() else 'cpu')
        elif isinstance(model_device, str):
            self.model_device = torch.device(model_device)
        else:
            self.model_device = model_device
        
        self.y_dim = y_dim
        self.n_batches = n_batches
        self.n_conditions = n_conditions
        self.n_subjects = n_subjects
        self.n_cell_types = n_cell_types
        self.effect_type = effect_type
        
        self.p_z_w = p_effect_z_w_class(model_device=self.model_device)
        self.p_z_w_coef = p_effect_z_w_coef_class(model_device=self.model_device,
                                                  y_dim=y_dim,
                                                  n_batches=n_batches,
                                                  n_conditions=n_conditions,
                                                  n_subjects=n_subjects,
                                                  n_cell_types=n_cell_types)
        self.q_z_w_coef = q_effect_z_w_coef_class(model_device=self.model_device,
                                                  y_dim=y_dim,
                                                  n_batches=n_batches,
                                                  n_conditions=n_conditions,
                                                  n_subjects=n_subjects,
                                                  n_cell_types=n_cell_types)
        self.p_z_w_hyper_coef = p_effect_z_w_hyper_coef_class(model_device=self.model_device,
                                                              y_dim=y_dim)
        self.q_z_w_hyper_coef = q_effect_z_w_hyper_coef_class(model_device=self.model_device,
                                                              y_dim=y_dim,
                                                              n_subjects=n_subjects)
        
        
    def _update_distributions(self, 
                              z_w: torch.tensor,
                              one_hot_index: torch.tensor,
                              FB: torch.tensor,
                              FC: torch.tensor,
                              RS: torch.tensor,
                              mode="training"):
        with torch.set_grad_enabled(mode=="training"):
            # To generate all the distributions, we start from
            # the posteriors 
            # The steps are almost always: generate the distributions
            # Then sample from the distributions 
            ####################################
            # Q
            ################
            q_z_w_coef_dict = self.q_z_w_coef()
            z_w_coef_samples = self.q_z_w_coef.get_samples()

            q_z_w_hyper_coef_dict = self.q_z_w_hyper_coef()
            z_w_hyper_coef_samples = self.q_z_w_hyper_coef.get_samples()
            
            ####################################
            # P
            ##############
            p_z_w_hyper_coef_dict = self.p_z_w_hyper_coef()
            
            p_z_w_coef_dict = self.p_z_w_coef(loc_eta_sample=z_w_hyper_coef_samples['loc_eta'],
                                              scale_eta_sample=z_w_hyper_coef_samples['scale_eta'])
            
            p_z_w_dict = self.p_z_w(one_hot_index=one_hot_index,
                                    FB=FB,
                                    FC=FC,
                                    RS=RS,
                                    theta_loc_sample=z_w_coef_samples['theta_loc'],
                                    alpha_loc_sample=z_w_coef_samples['alpha_loc'],
                                    beta_loc_sample=z_w_coef_samples['beta_loc'],
                                    gamma_loc_sample=z_w_coef_samples['gamma_loc'],
                                    alpha_theta_loc_sample=z_w_coef_samples['alpha_theta_loc'],
                                    beta_theta_loc_sample=z_w_coef_samples['beta_theta_loc'],
                                    theta_scale_sample=z_w_coef_samples['theta_scale'],
                                    alpha_scale_sample=z_w_coef_samples['alpha_scale'],
                                    beta_scale_sample=z_w_coef_samples['beta_scale'],
                                    gamma_scale_sample=z_w_coef_samples['gamma_scale'],
                                    alpha_theta_scale_sample=z_w_coef_samples['alpha_theta_scale'],
                                    beta_theta_scale_sample=z_w_coef_samples['beta_theta_scale'])
            log_likelihood = p_z_w_dict['z_w'].log_prob(z_w)
            
            return {
                "q_z_w_hyper_coef_dict": q_z_w_hyper_coef_dict,
                "p_z_w_hyper_coef_dict": p_z_w_hyper_coef_dict,
                "q_z_w_coef_dict": q_z_w_coef_dict,
                "p_z_w_coef_dict": p_z_w_coef_dict,
                "p_z_w_dict": p_z_w_dict,
                "log_likelihood": log_likelihood
            }
    
    def compute_loss(self,
                     distribution_dict: dict,
                     show_details: bool=False):
        reconstruction_error = distribution_dict['log_likelihood'].sum()
        
        kl_hyper_coef = 0
        for dist in distribution_dict['q_z_w_hyper_coef_dict']:
            kl_hyper_coef += kl_divergence(distribution_dict['q_z_w_hyper_coef_dict'][dist],
                                           distribution_dict['p_z_w_hyper_coef_dict'][dist]).sum()
        
        kl_coef = 0
        for dist in distribution_dict['q_z_w_coef_dict']:
            kl_coef += kl_divergence(distribution_dict['q_z_w_coef_dict'][dist],
                                    distribution_dict['p_z_w_coef_dict'][dist]).sum()
        
        elbo = reconstruction_error - kl_coef - kl_hyper_coef
        if show_details:
            print("="*25)
            print("likelihood is {}, kl_coef is {}, kl_hyper_coef is {}".format(reconstruction_error,
                                                                                kl_coef,
                                                                                kl_hyper_coef))
            print("="*25)
        return -elbo




# PyTorch 
import torch 
import torch.nn as nn 

# Typing 
from typing import Union

class component_base_class(nn.Module):
    def __init__(self,
                 stage_to_change: Union[str, list, tuple],
                 distribution_info: dict) -> None:
        """The basic class for all model components 

        Parameters
        ----------
        stage_to_change : Union[str, list, tuple]
            Training stages at which the weights can change 
            Namely, requires_grad=True
        distribution_info : dict
            A dictionary whose keys are the variable names. 
            Most of the time the corresponding values are simply None. 
            However for inferring different effects, the values are information 
            on shape of the distributions
        """
        super().__init__()
        # We first convert stage_to_change to a set 
        if isinstance(stage_to_change, str):
            stage_to_change = set([stage_to_change])
        else:
            stage_to_change = set(stage_to_change)
        
        assert stage_to_change <= set(["pretrain", "dimension reduction", "clustering", "abundance effect estimation", "expression effect estimation"]), "Illegal stage..."
        
        self.stage_to_change = stage_to_change
        
        has_extra_info = False
        # distribution_dict will store the actual distributions 
        self.distribution_dict = {}
        for dist in distribution_info:
            if distribution_info[dist] is not None:
                has_extra_info = True
            self.distribution_dict[dist] = None 
        # distribution_info_dict stores the information on the shape of the distributions 
        self.distribution_info_dict = {}
        if has_extra_info:
            for dist in distribution_info:
                self.distribution_info_dict[dist] = distribution_info[dist]
        # samples is a dictionary containing temporary samples 
        # generated from the distributions 
        self.samples = {dist: None for dist in self.distribution_dict}
        
    def _update_stage(self,
                      stage: str) -> None:
        """Update requires_grad according to the current stage 

        Parameters
        ----------
        stage : str
            The current stage 
        """
        # If the current stage is in stage_to_change
        # then the parameters will require gradients 
        for param in self.parameters():
            param.requires_grad = (stage in self.stage_to_change)
    
    def _update_distributions(self) -> None:
        """Defines how distributions are updated 

        Raises
        ------
        NotImplementedError
            This should be individually specified for each model component 
        """
        raise NotImplementedError
     
    def get_samples(self,
                    get_mean: bool=False) -> dict:
        """Generate a sample from each distribution

        Parameters
        ----------
        get_mean : bool, optional
            Whether or not to get the mean parameter of the distribution instead of 
            a random sample, by default False

        Returns
        -------
        dict
            A dictionary containing a random sample from each distribution
        """
        # We won't check if .mean or .sample is a method 
        # If not, PyTorch will throw an error 
        for dist in self.distribution_dict:
            if get_mean:
                self.samples[dist] = self.distribution_dict[dist].mean 
            else:
                if "rsample" in dir(self.distribution_dict[dist]):
                    self.samples[dist] = self.distribution_dict[dist].rsample()
                else:
                    self.samples[dist] = self.distribution_dict[dist].sample()
        return self.samples   
    
    def forward(self,
                **kwargs) -> dict:
        """Forward pass of the component 

        Returns
        -------
        dict
            A dictionary containing all the updated distributions 
        """
        self._update_distributions(**kwargs)
        return self.distribution_dict
 
class model_base_class(nn.Module):
    def __init__(self) -> None:
        """The basic class for all model classes
        
        A model class will contain one or several component objects 
        """
        super().__init__()

    def _update_stage(self,
                      stage: str) -> None:
        """Recursively update stages for all components contained within

        Parameters
        ----------
        stage : str
            The current training stage 
        """
        self.current_stage = stage
        for child in self.children():
            child._update_stage(stage=self.current_stage)
    
    def _update_distributions(self) -> dict:
        """Defines how distributions are updated 

        Returns
        -------
        dict
            A dictionary containing dictionaries of the updated distributions 
            and other information needed for computing loss

        Raises
        ------
        NotImplementedError
            This should be individually specified for each model 
        """
        raise NotImplementedError

    def forward(self,
                **kwargs) -> dict:
        """Forward pass of the model 

        Returns
        -------
        dict
            A dictionary containing dictionaries of the updated distributions 
            and other information needed for computing loss 
        """
        distribution_dict = self._update_distributions(**kwargs)
        return distribution_dict   
    
    def compute_loss(self) -> torch.tensor:
        """Defines how loss should be computed 

        Returns
        -------
        torch.tensor
            A tensor of loss

        Raises
        ------
        NotImplementedError
            This should be individually specified for each model
        """
        raise NotImplementedError
    
    


import torch 
import torch.nn as nn 
import torch.nn.functional as F
from torch.distributions import Normal, Independent
from pyro.distributions import Delta 

from CytoOne.base_class import component_base_class


class p_effect_z_w_coef_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 y_dim: int,
                 n_batches: int,
                 n_conditions: int,
                 n_subjects: int,
                 n_cell_types: int): 
        super().__init__(stage_to_change="expression effect estimation",
                         distribution_info={"alpha_loc": (n_batches, y_dim),
                                            "beta_loc": (n_conditions, y_dim),
                                            "gamma_loc": (n_subjects, y_dim),
                                            "theta_loc": (n_cell_types, y_dim),
                                            "alpha_theta_loc": (n_batches*n_cell_types, y_dim),
                                            "beta_theta_loc": (n_conditions*n_cell_types, y_dim),
                                            "alpha_scale": (n_batches, y_dim),
                                            "beta_scale": (n_conditions, y_dim),
                                            "gamma_scale": (n_subjects, y_dim),
                                            "theta_scale": (n_cell_types, y_dim),
                                            "alpha_theta_scale": (n_batches*n_cell_types, y_dim),
                                            "beta_theta_scale": (n_conditions*n_cell_types, y_dim)})
        self.model_device = model_device
        
        self.y_dim = y_dim
        self.n_batches = n_batches
        self.n_conditions = n_conditions
        self.n_subjects = n_subjects
        self.n_cell_types = n_cell_types
        
    def _update_distributions(self,
                              loc_eta_sample: torch.tensor,
                              scale_eta_sample: torch.tensor) -> None:
        
        gamma_loc_scale = torch.sqrt(torch.exp(loc_eta_sample))
        gamma_scale_scale = torch.sqrt(torch.exp(scale_eta_sample))
        
        for dist in self.distribution_dict:
            if (self.distribution_info_dict[dist][0] == 1) or \
               (("_theta_" in dist) and (self.distribution_info_dict[dist][0] == self.n_cell_types)) or \
               (("_theta_" in dist) and (self.n_cell_types == 1)):
                self.distribution_dict[dist] = Independent(Delta(v=torch.zeros(self.distribution_info_dict[dist],
                                                                               device=self.model_device),
                                                                 log_density=torch.ones(self.distribution_info_dict[dist],
                                                                                        device=self.model_device)),
                                                           reinterpreted_batch_ndims=len(self.distribution_info_dict[dist]))
            else:
                if dist == "gamma_loc":
                    self.distribution_dict[dist] = Independent(Normal(loc=torch.zeros(self.distribution_info_dict[dist],
                                                                                      device=self.model_device),
                                                                      scale=torch.ones(self.distribution_info_dict[dist],
                                                                                       device=self.model_device)*gamma_loc_scale),
                                                               reinterpreted_batch_ndims=len(self.distribution_info_dict[dist])) 
                elif dist == "gamma_scale":
                    self.distribution_dict[dist] = Independent(Normal(loc=torch.zeros(self.distribution_info_dict[dist],
                                                                                      device=self.model_device),
                                                                      scale=torch.ones(self.distribution_info_dict[dist],
                                                                                       device=self.model_device)*gamma_scale_scale),
                                                               reinterpreted_batch_ndims=len(self.distribution_info_dict[dist])) 
                else:
                    self.distribution_dict[dist] = Independent(Normal(loc=torch.zeros(self.distribution_info_dict[dist],
                                                                                      device=self.model_device),
                                                                      scale=torch.ones(self.distribution_info_dict[dist],
                                                                                       device=self.model_device)),
                                                               reinterpreted_batch_ndims=len(self.distribution_info_dict[dist]))
        

class q_effect_z_w_coef_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 y_dim: int,
                 n_batches: int,
                 n_conditions: int,
                 n_subjects: int,
                 n_cell_types: int):
        super().__init__(stage_to_change = "expression effect estimation",
                         distribution_info={"alpha_loc": (n_batches, y_dim),
                                            "beta_loc": (n_conditions, y_dim),
                                            "gamma_loc": (n_subjects, y_dim),
                                            "theta_loc": (n_cell_types, y_dim),
                                            "alpha_theta_loc": (n_batches*n_cell_types, y_dim),
                                            "beta_theta_loc": (n_conditions*n_cell_types, y_dim),
                                            "alpha_scale": (n_batches, y_dim),
                                            "beta_scale": (n_conditions, y_dim),
                                            "gamma_scale": (n_subjects, y_dim),
                                            "theta_scale": (n_cell_types, y_dim),
                                            "alpha_theta_scale": (n_batches*n_cell_types, y_dim),
                                            "beta_theta_scale": (n_conditions*n_cell_types, y_dim)})
        self.model_device = model_device
        self.y_dim = y_dim
        self.n_batches = n_batches
        self.n_conditions = n_conditions
        self.n_subjects = n_subjects
        self.n_cell_types = n_cell_types
        
        self.parameter_dict = nn.ParameterDict({})
        
        for dist in self.distribution_dict:
            if (self.distribution_info_dict[dist][0] > 1) and \
               (("_theta_" not in dist) or (self.distribution_info_dict[dist][0] != n_cell_types)) and \
               (("_theta_" not in dist) or (self.n_cell_types > 1)):
                self.parameter_dict.update({dist+"_loc": nn.Parameter(torch.randn(self.distribution_info_dict[dist]),
                                                                            requires_grad=True),
                                            dist+"_scale": nn.Parameter(torch.randn(self.distribution_info_dict[dist]),
                                                                            requires_grad=True)})
    
    def _update_distributions(self):
        for dist in self.distribution_dict:
            if (self.distribution_info_dict[dist][0] == 1) or \
               (("_theta_" in dist) and (self.distribution_info_dict[dist][0] == self.n_cell_types)) or \
               (("_theta_" in dist) and (self.n_cell_types == 1)):
                self.distribution_dict[dist] = Independent(Delta(v=torch.zeros(self.distribution_info_dict[dist],
                                                                               device=self.model_device),
                                                                 log_density=torch.ones(self.distribution_info_dict[dist],
                                                                                        device=self.model_device)),
                                                           reinterpreted_batch_ndims=len(self.distribution_info_dict[dist]))
            else:
                self.distribution_dict[dist] = Independent(Normal(loc=self.parameter_dict[dist+'_loc'],
                                                                  scale=F.softplus(self.parameter_dict[dist+'_scale'], beta=1)+0.00001),
                                                            reinterpreted_batch_ndims=len(self.distribution_info_dict[dist]))


class p_effect_z_w_hyper_coef_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 y_dim: int):
        super().__init__(stage_to_change = "expression effect estimation",
                         distribution_info={"loc_eta": (1, y_dim),
                                            "scale_eta": (1, y_dim)})
        self.model_device = model_device
        self.y_dim = y_dim
        
    def _update_distributions(self):
        for dist in self.distribution_dict:
            self.distribution_dict[dist] = Independent(Normal(loc=torch.zeros(self.distribution_info_dict[dist],
                                                                              device=self.model_device),
                                                              scale=torch.ones(self.distribution_info_dict[dist],
                                                                               device=self.model_device)),
                                                       reinterpreted_batch_ndims=2)
        
        
class q_effect_z_w_hyper_coef_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 y_dim: int,
                 n_subjects: int):
        super().__init__(stage_to_change = "expression effect estimation",
                         distribution_info={"loc_eta": (1, y_dim),
                                            "scale_eta": (1, y_dim)})
        self.model_device = model_device
        self.y_dim = y_dim
        self.n_subjects = n_subjects
        self.parameter_dict = nn.ParameterDict({})
        
        for dist in self.distribution_dict:
            if n_subjects > 1:
                self.parameter_dict.update({dist+"_loc": nn.Parameter(torch.randn(self.distribution_info_dict[dist]),
                                                                            requires_grad=True),
                                            dist+"_scale": nn.Parameter(torch.randn(self.distribution_info_dict[dist]),
                                                                            requires_grad=True)})
        
    def _update_distributions(self):
        for dist in self.distribution_dict:
            if self.n_subjects == 1:
                self.distribution_dict[dist] = Independent(Normal(loc=torch.zeros(self.distribution_info_dict[dist],
                                                                                  device=self.model_device),
                                                                  scale=torch.ones(self.distribution_info_dict[dist],
                                                                                   device=self.model_device)),
                                                           reinterpreted_batch_ndims=2)
            else:
                self.distribution_dict[dist] = Independent(Normal(loc=self.parameter_dict[dist+'_loc'],
                                                                  scale=F.softplus(self.parameter_dict[dist+'_scale'], beta=1)+0.00001),
                                                            reinterpreted_batch_ndims=len(self.distribution_info_dict[dist]))
    


import torch 
import torch.nn as nn 
import torch.nn.functional as F
from torch.distributions import Categorical, Independent
from pyro.distributions import Delta 

from CytoOne.base_class import component_base_class

import numpy as np 
from collections import OrderedDict


class p_effect_pi_class(component_base_class):
    def __init__(self,
                 model_device: torch.device):
        super().__init__(stage_to_change="abundance effect estimation",
                         distribution_info={"pi": None})
        self.model_device = model_device
        
    def _update_distributions(self, 
                              FC: torch.tensor,
                              RS: torch.tensor, 
                              beta_sample: torch.tensor,
                              gamma_sample: torch.tensor):
        
        logits = torch.matmul(FC, beta_sample) + torch.matmul(RS, gamma_sample)
        probs = F.softmax(logits, dim=1)
        self.distribution_dict['pi'] = Independent(Categorical(probs=probs),
                                                   reinterpreted_batch_ndims=0)


class p_pi_class(component_base_class):
    def __init__(self, 
                 model_device: torch.device,
                 n_cell_types: int,
                 n_conditions: int=1,
                 n_subjects: int=1) -> None:
        super().__init__(stage_to_change="clustering", 
                         distribution_info={"pi": None})
        self.model_device = model_device
        extra_n_dim = np.sum([n for n in [n_conditions, n_subjects] if n>1], dtype=int)
        if extra_n_dim > 1:
            self.in_dim = extra_n_dim
        else:
            self.in_dim = 1
        self.logit_mapping = nn.Sequential(OrderedDict([
            ('fc1', nn.Linear(in_features=self.in_dim,
                              out_features=n_cell_types,
                              bias=False))
        ]))
        
    def _update_distributions(self, 
                              FC: torch.tensor,
                              RS: torch.tensor):
        if self.in_dim == 1:
            effects = torch.ones((FC.shape[0], 1),
                                 device=self.model_device) 
        else:
            extra_effect_list = [m for m in [FC, RS] if m.shape[1] > 1]
            effects = torch.cat(extra_effect_list, dim=1)
            
        logits = self.logit_mapping(effects) 
        probs = F.softmax(logits, dim=1)
        self.distribution_dict['pi'] = Independent(Categorical(probs=probs),
                                                   reinterpreted_batch_ndims=0)    
        

class q_pi_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 x_dim: int,
                 n_cell_types: int,
                 vq_vae_weight: float) -> None:
        super().__init__(stage_to_change="clustering", 
                         distribution_info={"pi": None,
                                            "embedding": None,
                                            "vq_loss": None})
        self.model_device = model_device
        self.vq_vae_weight = vq_vae_weight
        self.cell_embeddings = nn.Embedding(num_embeddings=n_cell_types,
                                            embedding_dim=x_dim)
                    
    def _update_distributions(self, x):
        # x**2 should be an N * x_dim tensor
        # After torch.sum with keepdim=True,
        # the resulting tensor should be N * 1
        
        # The weight of cell_embeddings should be a 
        # P * x_dim tensor. After torch.sum with keepdim=False 
        # the resulting tensor should be P
        
        # Summing these two tensors up would by 
        # broadcasting give us an N * P tensor 
        
        distance = torch.sum(x**2, dim=1, keepdim=True) + \
                        torch.sum(self.cell_embeddings.weight**2, dim=1, keepdim=False) - \
                        2 * torch.matmul(x, self.cell_embeddings.weight.t())
        
        # Now for each row of X, we will find the index 
        # of the closest embedding 
        index = torch.argmin(distance, dim=1).unsqueeze(dim=1)
        
        # Then we convert this index to its one-hot encoding 
        one_hot_encoding = torch.zeros(distance.size(0), 
                                       distance.size(1))
        one_hot_encoding.scatter_(1, index, 1)
        
        quantized_latents = torch.matmul(one_hot_encoding, self.cell_embeddings.weight)
        
        # Compute VQ loss 
        commitment_loss = F.mse_loss(quantized_latents.detach(), x)
        embedding_loss = F.mse_loss(quantized_latents, x.detach())
        
        vq_loss = commitment_loss * self.vq_vae_weight  + embedding_loss
        
        quantized_latents = x + (quantized_latents-x).detach() 
        
        self.distribution_dict['pi'] = Independent(Delta(v=index.squeeze(1),
                                                         log_density=1), 
                                                         reinterpreted_batch_ndims=0)
        self.distribution_dict['embedding'] = quantized_latents
        self.distribution_dict['vq_loss'] = vq_loss
        self.distribution_dict['one_hot_encoding'] = one_hot_encoding



# Data IO 
import os 
import numpy as np 
import pandas as pd 

# PyTorch
import torch 
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader

# Typing 
from typing import Optional, Union


class dataset_class(Dataset):
    def __init__(self, 
                 model_device: torch.device,
                 df: pd.DataFrame,
                 meta_dict: Optional[dict]=None) -> None:
        """Dataset class for all subsequent models 
        
        
        Parameters
        ----------
        df : pd.DataFrame
            A curated CyTOF dataframe  
        meta_dict : Optional[dict], optional
            _description_, by default None
        """
        super().__init__()
        
        self.model_device = model_device 
        self.data_type = None
        
        if meta_dict is None:
            self.data_type = "pretrain"
            self.y = torch.tensor(df[['y']].values, dtype=torch.float32)
        else:
            self.data_type = "post-pretraining"
            self.y = torch.tensor(df[meta_dict['protein_col_names']].values, dtype=torch.float32)
            
            b = torch.tensor(df['batch_index'].values, dtype=torch.int64).unsqueeze(1)
            c = torch.tensor(df['condition_index'].values, dtype=torch.int64).unsqueeze(1)
            s = torch.tensor(df['subject_index'].values, dtype=torch.int64).unsqueeze(1)
            theta = torch.tensor(df['cell_type_index'].values, dtype=torch.int64).unsqueeze(1)
            
            self.FB = torch.zeros(meta_dict['N'], meta_dict['n_batches'])
            self.FC = torch.zeros(meta_dict['N'], meta_dict['n_conditions'])
            self.RS = torch.zeros(meta_dict['N'], meta_dict['n_subjects'])
            self.Theta = torch.zeros(meta_dict['N'], meta_dict['n_types'])
            
            self.FB.scatter_(1, b, 1)
            self.FC.scatter_(1, c, 1)
            self.RS.scatter_(1, s, 1)
            self.Theta.scatter_(1, theta, 1)
    
    def __len__(self) -> int:
        return self.y.shape[0]
    
    def __getitem__(self, index) -> dict:
        if self.data_type == "pretrain":
            return {"y": self.y[index, :].to(self.model_device)}
        else:
            return {"y": self.y[index, :].to(self.model_device), 
                    "FB": self.FB[index, :].to(self.model_device),
                    "FC": self.FC[index, :].to(self.model_device), 
                    "RS": self.RS[index, :].to(self.model_device)}, {"Theta": self.Theta[index, :].to(self.model_device)}


def generate_dataloader(df: pd.DataFrame,
                        meta_dict: Optional[dict]=None,
                        model_device: Optional[Union[str, torch.device]]=None,
                        batch_size: int=256,
                        shuffle: bool=True) -> DataLoader:
    if model_device is None:
        model_device = torch.device(
            'cuda:0' if torch.cuda.is_available() else 'cpu')
    elif isinstance(model_device, str):
        model_device = torch.device(model_device)
    else:
        model_device = model_device
    
    dataset = dataset_class(model_device=model_device,
                            df=df,
                            meta_dict=meta_dict)
    return DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)


def training_loop(model,
                  optimizer: torch.optim.Optimizer,
                  scheduler, 
                  n_epoches: int,
                  train_dataloader: DataLoader,
                  val_dataloader: DataLoader,
                  models_dir: str,
                  model_name: str,
                  training_stage: str,
                  save_every_epoch: bool=False,
                  print_every_n_minibatch: int=10,
                  cytoone_model: Optional[torch.nn.Module]=None,
                  use_true_cell_types: bool=True,
                  show_details: bool=False) -> None:

    model._update_stage(stage=training_stage)
    
    best_val_loss = np.inf
    for epoch in range(1, n_epoches + 1):
        # First train on 
        model.train()
        for minibatch, (data, one_hot_index) in enumerate(train_dataloader):
            if training_stage in ["pretrain", "dimension reduction", "clustering"]:
                distribution_dict = model(**data) 
            else:
                (_, _, q_pi_dict, z_w_samples, _) = cytoone_model.get_posterior_samples(**data)
                if not use_true_cell_types:
                    one_hot_index = q_pi_dict['one_hot_encoding']
                z = z_w_samples['z']
                w = z_w_samples['w']
                if training_stage == "abundance effect estimation":
                    distribution_dict = model(one_hot_index=one_hot_index,
                                              FC=data["FC"],
                                              RS=data["RS"]) 
                elif training_stage == "expression effect estimation":
                    if model.effect_type == "expression":
                        distribution_dict = model(z_w=z,
                                                  one_hot_index=one_hot_index,
                                                  FB=data['FB'],
                                                  FC=data['FC'],
                                                  RS=data['RS']) 
                    elif model.effect_type == "inflation":
                        distribution_dict = model(z_w=w,
                                                  one_hot_index=one_hot_index,
                                                  FB=data['FB'],
                                                  FC=data['FC'],
                                                  RS=data['RS'])  
                    else: 
                        raise RuntimeError("Illegal training stage")  
                else: 
                    raise RuntimeError("Illegal training stage")  
                
            training_loss = model.compute_loss(distribution_dict=distribution_dict,
                                               show_details=show_details)

            optimizer.zero_grad()
            training_loss.backward()
            #Gradient Value Clipping
            nn.utils.clip_grad_norm_(model.parameters(), max_norm=1.0, norm_type=2)
            optimizer.step()
            
            if (epoch % 1 == 0) and (minibatch % print_every_n_minibatch == 0):
                print("Epoch {}, minibatch {}. Training loss is {}\n".format(epoch, minibatch, training_loss.item()))
                print("="*25)
                
        model.eval()
        total_val_loss = 0.0
        with torch.no_grad():
            for (data, one_hot_index) in val_dataloader:
                if training_stage in ["pretrain", "dimension reduction", "clustering"]:
                    val_distributions_dict = model(**data) 
                else:
                    (_, _, q_pi_dict, z_w_samples, _) = cytoone_model.get_posterior_samples(**data)
                    if not use_true_cell_types:
                        one_hot_index = q_pi_dict['one_hot_encoding']
                    z = z_w_samples['z']
                    w = z_w_samples['w']
                    if training_stage == "abundance effect estimation":
                        val_distributions_dict = model(one_hot_index=one_hot_index,
                                                       FC=data['FC'],
                                                       RS=data['RS']) 
                    elif training_stage == "expression effect estimation":
                        if model.effect_type == "expression":
                            val_distributions_dict = model(z_w=z,
                                                           one_hot_index=one_hot_index,
                                                           FB=data['FB'],
                                                           FC=data['FC'],
                                                           RS=data['RS']) 
                        elif model.effect_type == "inflation":
                            val_distributions_dict = model(z_w=w,
                                                           one_hot_index=one_hot_index,
                                                           FB=data['FB'],
                                                           FC=data['FC'],
                                                           RS=data['RS'])  
                        else: 
                            raise RuntimeError("Illegal training stage") 
                    else: 
                        raise RuntimeError("Illegal training stage")  
                validation_loss = model.compute_loss(distribution_dict=val_distributions_dict,
                                                     show_details=show_details)
                total_val_loss += validation_loss.item()
            
            if save_every_epoch:
                torch.save({
                    'epoch': epoch,
                    'model_state_dict': model.state_dict(),
                    'optimizer_state_dict': optimizer.state_dict()
                }, os.path.join(models_dir, model_name + "_" + str(epoch) + ".pth"))
            else:
                if (total_val_loss/len(val_dataloader)) < best_val_loss:
                    best_val_loss = validation_loss.item()
                    torch.save({
                        'epoch': epoch,
                        'model_state_dict': model.state_dict(),
                        'optimizer_state_dict': optimizer.state_dict()
                    }, os.path.join(models_dir, model_name + ".pth"))
        
        if scheduler is not None:
            scheduler.step()
        
        if epoch % 1 == 0:
            print('Epoch {} Validation loss {}\n'.format(epoch, total_val_loss/len(val_dataloader)))
            print("="*25) 




# PyTorch 
import torch 
import torch.nn as nn 
import torch.nn.functional as F
from torch.distributions import Normal, Independent

# Data manipulation 
import numpy as np 

# For module construction 
from collections import OrderedDict

# Typing 
from typing import Optional

# Basic class 
from CytoOne.base_class import component_base_class


class p_pretrain_z_class(component_base_class):
    def __init__(self,
                 model_device: torch.device) -> None:
        super().__init__(stage_to_change='pretrain', 
                         distribution_info={'z': None})
        self.model_device = model_device
        self.loc = torch.tensor(0, 
                                dtype=torch.float32,
                                device=self.model_device)
        self.scale = torch.tensor(5, 
                                  dtype=torch.float32,
                                  device=self.model_device)

    def _update_distributions(self):
        self.distribution_dict['z'] = Independent(Normal(loc=self.loc,
                                                         scale=self.scale), 
                                                  reinterpreted_batch_ndims=0)  

   
class p_pretrain_w_class(component_base_class):
    def __init__(self,
                 model_device: torch.device) -> None:
        super().__init__(stage_to_change='pretrain', 
                         distribution_info={'w': None})
        self.model_device = model_device
        self.loc = torch.tensor(0, 
                                dtype=torch.float32,
                                device=self.model_device)
        self.scale = torch.tensor(5, 
                                  dtype=torch.float32,
                                  device=self.model_device)

    def _update_distributions(self):
        self.distribution_dict['w'] = Independent(Normal(loc=self.loc,
                                                         scale=self.scale), 
                                                  reinterpreted_batch_ndims=0)   


class p_effect_z_w_class(component_base_class):
    def __init__(self,
                 model_device: torch.device):
        super().__init__(stage_to_change="expression effect estimation",
                         distribution_info={'z_w': None})
        self.model_device = model_device
        
    def _update_distributions(self,
                              one_hot_index: torch.tensor,
                              FB: torch.tensor,
                              FC: torch.tensor,
                              RS: torch.tensor, 
                              theta_loc_sample: torch.tensor,
                              alpha_loc_sample: torch.tensor,
                              beta_loc_sample: torch.tensor,
                              gamma_loc_sample: torch.tensor,
                              alpha_theta_loc_sample: torch.tensor,
                              beta_theta_loc_sample: torch.tensor,
                              theta_scale_sample: torch.tensor,
                              alpha_scale_sample: torch.tensor,
                              beta_scale_sample: torch.tensor,
                              gamma_scale_sample: torch.tensor,
                              alpha_theta_scale_sample: torch.tensor,
                              beta_theta_scale_sample: torch.tensor):
        FB_theta = torch.cat([one_hot_index[:, [c]] * FB for c in range(one_hot_index.shape[1])],
                             dim=1)
        FC_theta = torch.cat([one_hot_index[:, [c]] * FC for c in range(one_hot_index.shape[1])],
                             dim=1)
        
        loc = torch.matmul(one_hot_index, theta_loc_sample) + \
              torch.matmul(FB, alpha_loc_sample) + \
              torch.matmul(FC, beta_loc_sample) + \
              torch.matmul(RS, gamma_loc_sample) + \
              torch.matmul(FB_theta, alpha_theta_loc_sample) + \
              torch.matmul(FC_theta, beta_theta_loc_sample) 
              
        log_scale = torch.matmul(one_hot_index, theta_scale_sample) + \
                    torch.matmul(FB, alpha_scale_sample) + \
                    torch.matmul(FC, beta_scale_sample) + \
                    torch.matmul(RS, gamma_scale_sample) + \
                    torch.matmul(FB_theta, alpha_theta_scale_sample) + \
                    torch.matmul(FC_theta, beta_theta_scale_sample) 

        self.distribution_dict['z_w'] = Independent(Normal(loc=loc,
                                                           scale=F.softplus(log_scale, beta=1) + 0.00001),
                                                  reinterpreted_batch_ndims=1)
        


class p_z_w_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 y_dim: int,
                 x_dim: int,
                 n_batches: int=1,
                 n_conditions: int=1,
                 n_subjects: int=1) -> None:
        super().__init__(stage_to_change="dimension reduction",
                         distribution_info={"z": None,
                                            "w": None})
        self.model_device = model_device
        extra_n_dim = np.sum([n for n in [n_batches, n_conditions, n_subjects] if n>1], dtype=int)
        self.mu_z_mapping = nn.Sequential(OrderedDict([
            ('fc1', nn.Linear(in_features=x_dim + extra_n_dim,
                                out_features=128,
                                bias=True)),
            ('fc1_bn', nn.BatchNorm1d(128)),
            ('fc1_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc2', nn.Linear(in_features=128,
                                out_features=256,
                                bias=True)),
            ('fc2_bn', nn.BatchNorm1d(256)),
            ('fc2_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc3', nn.Linear(in_features=256,
                                out_features=512,
                                bias=True)),
            ('fc3_bn', nn.BatchNorm1d(512)),
            ('fc3_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc4', nn.Linear(in_features=512,
                                out_features=y_dim,
                                bias=True))
        ]))
        
        self.log_Sigma_z_mapping = nn.Sequential(OrderedDict([
            ('fc1', nn.Linear(in_features=x_dim + extra_n_dim,
                                out_features=128,
                                bias=True)),
            ('fc1_bn', nn.BatchNorm1d(128)),
            ('fc1_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc2', nn.Linear(in_features=128,
                                out_features=256,
                                bias=True)),
            ('fc2_bn', nn.BatchNorm1d(256)),
            ('fc2_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc3', nn.Linear(in_features=256,
                                out_features=512,
                                bias=True)),
            ('fc3_bn', nn.BatchNorm1d(512)),
            ('fc3_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc4', nn.Linear(in_features=512,
                                out_features=y_dim,
                                bias=True))
        ]))
        
        self.mu_w_mapping = nn.Sequential(OrderedDict([
            ('fc1', nn.Linear(in_features=x_dim + extra_n_dim,
                            out_features=128,
                            bias=True)),
            ('fc1_bn', nn.BatchNorm1d(128)),
            ('fc1_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc2', nn.Linear(in_features=128,
                                out_features=256,
                                bias=True)),
            ('fc2_bn', nn.BatchNorm1d(256)),
            ('fc2_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc3', nn.Linear(in_features=256,
                                out_features=512,
                                bias=True)),
            ('fc3_bn', nn.BatchNorm1d(512)),
            ('fc3_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc4', nn.Linear(in_features=512,
                                out_features=y_dim,
                                bias=True))
        ]))
        
        self.log_Sigma_w_mapping = nn.Sequential(OrderedDict([
            ('fc1', nn.Linear(in_features=x_dim + extra_n_dim,
                                out_features=128,
                                bias=True)),
            ('fc1_bn', nn.BatchNorm1d(128)),
            ('fc1_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc2', nn.Linear(in_features=128,
                                out_features=256,
                                bias=True)),
            ('fc2_bn', nn.BatchNorm1d(256)),
            ('fc2_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc3', nn.Linear(in_features=256,
                                out_features=512,
                                bias=True)),
            ('fc3_bn', nn.BatchNorm1d(512)),
            ('fc3_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc4', nn.Linear(in_features=512,
                                out_features=y_dim,
                                bias=True))
        ]))
                    
    def _update_distributions(self,
                              x: torch.tensor,
                              FB: torch.tensor,
                              FC: torch.tensor,
                              RS: torch.tensor):
        
        effect_list = [m for m in [FB, FC, RS] if m.shape[1] > 1]
        
        x_with_effect = torch.cat([x]+ effect_list, dim=1)
        
        z_loc = self.mu_z_mapping(x_with_effect) 
        # z_scale = torch.tensor(0.1, dtype=torch.float32)
        z_scale = F.softplus(self.log_Sigma_z_mapping(x_with_effect), beta=1) + 0.00001
        w_loc = self.mu_w_mapping(x_with_effect) 
        # w_scale = torch.tensor(0.1, dtype=torch.float32)
        w_scale = F.softplus(self.log_Sigma_w_mapping(x_with_effect), beta=1) + 0.00001
        self.distribution_dict['z'] = Independent(Normal(loc=z_loc,
                                                         scale=z_scale),
                                                    reinterpreted_batch_ndims=1)
        self.distribution_dict['w'] = Independent(Normal(loc=w_loc,
                                                         scale=w_scale),
                                                    reinterpreted_batch_ndims=1)


class q_pretrain_z_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 y_dim: int=1) -> None:
        super().__init__(stage_to_change='pretrain', 
                         distribution_info={'z': None})
        self.model_device = model_device
        self.y_dim = y_dim
        self.reinterpreted_batch_ndims = 0
        if self.y_dim > 1:
            self.reinterpreted_batch_ndims = 1
            
        self.mu_z_mapping = nn.Sequential(OrderedDict([
            ('fc1', nn.Linear(in_features=1,
                                out_features=16,
                                bias=True)),
            ('fc1_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc2', nn.Linear(in_features=16,
                                out_features=8,
                                bias=True)),
            ('fc2_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc3', nn.Linear(in_features=8,
                                out_features=4,
                                bias=True)),
            ('fc3_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc4', nn.Linear(in_features=4,
                                out_features=1,
                                bias=True))
        ]))
        
        self.log_Sigma_z_mapping = nn.Sequential(OrderedDict([
            ('fc1', nn.Linear(in_features=1,
                                out_features=16,
                                bias=True)),
            ('fc1_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc2', nn.Linear(in_features=16,
                                out_features=8,
                                bias=True)),
            ('fc2_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc3', nn.Linear(in_features=8,
                                out_features=4,
                                bias=True)),
            ('fc3_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc4', nn.Linear(in_features=4,
                                out_features=1,
                                bias=True))
        ]))
    
    def _update_distributions(self,
                              y: torch.tensor):        
        flattened_y = y.view((-1, 1))
        
        loc = self.mu_z_mapping(flattened_y)
        scale = F.softplus(self.log_Sigma_z_mapping(flattened_y)) + 0.00001
        
        # reshape z_loc and z_scale 
        unflattened_loc = loc.view((-1, self.y_dim))
        unflattened_scale = scale.view((-1, self.y_dim))
        
        self.distribution_dict['z'] = Independent(Normal(loc=unflattened_loc,
                                                         scale=unflattened_scale),
                                                  reinterpreted_batch_ndims=self.reinterpreted_batch_ndims)
        
    
class q_pretrain_w_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 y_dim: int=1) -> None:
        super().__init__(stage_to_change='pretrain', 
                         distribution_info={'w': None})
        self.model_device = model_device
        self.y_dim = y_dim
        self.reinterpreted_batch_ndims = 0
        if self.y_dim > 1:
            self.reinterpreted_batch_ndims = 1
        
        self.mu_w_mapping = nn.Sequential(OrderedDict([
            ('fc1', nn.Linear(in_features=1,
                                out_features=16,
                                bias=True)),
            ('fc1_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc2', nn.Linear(in_features=16,
                                out_features=8,
                                bias=True)),
            ('fc2_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc3', nn.Linear(in_features=8,
                                out_features=4,
                                bias=True)),
            ('fc3_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc4', nn.Linear(in_features=4,
                                out_features=1,
                                bias=True))
        ]))
        
        self.log_Sigma_w_mapping = nn.Sequential(OrderedDict([
            ('fc1', nn.Linear(in_features=1,
                                out_features=16,
                                bias=True)),
            ('fc1_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc2', nn.Linear(in_features=16,
                                out_features=8,
                                bias=True)),
            ('fc2_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc3', nn.Linear(in_features=8,
                                out_features=4,
                                bias=True)),
            ('fc3_relu', nn.LeakyReLU(negative_slope=0.2)),
            ('fc4', nn.Linear(in_features=4,
                                out_features=1,
                                bias=True))
        ]))
    
    def _update_distributions(self,
                              y: torch.tensor):
        flattened_y = y.view((-1, 1))
        
        loc = self.mu_w_mapping(flattened_y)
        scale = F.softplus(self.log_Sigma_w_mapping(flattened_y)) + 0.00001
        
        # reshape z_loc and z_scale 
        unflattened_loc = loc.view((-1, self.y_dim))
        unflattened_scale = scale.view((-1, self.y_dim))
        
        self.distribution_dict['w'] = Independent(Normal(loc=unflattened_loc,
                                                         scale=unflattened_scale),
                                                  reinterpreted_batch_ndims=self.reinterpreted_batch_ndims)
    

class q_z_w_class(component_base_class):
    def __init__(self,
                 model_device: torch.device,
                 y_dim: int) -> None:
        super().__init__(stage_to_change="pretrain",
                         distribution_info={"z": None,
                                            "w": None})
        self.model_device = model_device
        self.y_dim = y_dim
        self.q_z = q_pretrain_z_class(model_device=self.model_device,
                                      y_dim=y_dim)
        self.q_w = q_pretrain_w_class(model_device=self.model_device,
                                      y_dim=y_dim)
    
    def load_pretrained_model(self,
                              model_check_point_path: str) -> None:
        ckpt = torch.load(model_check_point_path,
                          map_location=self.model_device)
        model_dict = self.state_dict()

        # 1. filter out unnecessary keys
        # pretrained_dict = {k.replace("q_z.", "").replace("q_w.", ""): v for k, v in ckpt['model_state_dict'].items() if (k.replace("q_z.", "") in model_dict) or (k.replace("q_w.", "") in model_dict)}
        pretrained_dict = {k: v for k, v in ckpt['model_state_dict'].items() if (k in model_dict)}
        # 2. overwrite entries in the existing state dict
        model_dict.update(pretrained_dict)
        self.load_state_dict(pretrained_dict)
        # 3. Then, we freeze the parameters 
        for param in self.parameters():
            param.requires_grad = False
        
    def _update_distributions(self,
                              y: torch.tensor):
        self.distribution_dict['z'] = self.q_z(y=y)['z']
        self.distribution_dict['w'] = self.q_w(y=y)['w']





from setuptools import setup 
import sys 
import os 
import shutil
import distutils.cmd


VERSION = "0.0.1"

class PypiCommand(distutils.cmd.Command):
    
    description = "Build and upload for PyPI."
    user_options = []
    
    def initialize_options(self):
        pass
    
    
    def finalize_options(self):
        pass
    
    
    def run(self):
        try:
            shutil.rmtree("dist/")
        except FileNotFoundError:
            pass
        
        wheel_file = "CytoOne-{}-py3-none-any.whl".format(VERSION)
        tar_file = "CytoOne-{}.tar.gz".format(VERSION)
        
        os.system("{} setup.py sdist bdist_wheel".format(sys.executable))
        os.system("twine upload dist/{} dist/{}".format(wheel_file, tar_file))
    
    
# class CondaCommand(distutils.cmd.Command):
    
#     description = "Build and upload for conda."
#     user_options = []
    
#     def initialize_options(self):
#         pass
    
    
#     def finalize_options(self):
#         pass
    
    
#     def run(self):
#         try:
#             shutil.rmtree("dist_conda/")
#         except FileNotFoundError:
#             pass
#         os.system("conda build . --output-folder dist_conda/")
#         os.system("anaconda upload ./dist_conda/noarch/pMTnet_Omni-{}-py_0.tar.bz2".format(VERSION))


setup(
    name="CytoOne",
    version=VERSION,
    description="A unified probabilistic framework for CyTOF data",
    author="Yuqiu Yang, Kevin Wang, Tao Wang, Xinlei (Sherry) Wang",
    author_email="yuqiuy@smu.edu, kevinwang@mail.smu.edu, Tao.Wang@UTSouthwestern.edu, xinlei.wang@uta.edu",
    long_description_content_type="text/markdown",
    long_description=open("README.md").read(),
    packages=["CytoOne"],
    python_requires=">=3.9,<3.11",
    install_requires=[
        "numpy==1.22.4",
        "pandas==1.5.2",
        "torch==1.13.1",
        "pyro-ppl==1.8.6",
        "matplotlib==3.7.1",
        "seaborn==0.13.0"
    ],
    test_requires=[
        'pytest==7.1.2',
        'coverage==6.3.2',
        'pytest-cov==3.0.0',
        'pytest-mock==3.10.0'
    ],
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Natural Language :: English",
        "Intended Audience :: Science/Research",
    ],
    cmdclass={
        "pypi": PypiCommand,
        # "conda": CondaCommand
    }
)




# PyTorch 
import torch 
import torch.nn as nn 
from torch.distributions.kl import kl_divergence

from CytoOne.pi_class import p_effect_pi_class
from CytoOne.pi_effect_coef_class import p_effect_pi_coef_class, p_effect_pi_hyper_coef_class, \
                                         q_effect_pi_coef_class, q_effect_pi_hyper_coef_class

from CytoOne.base_class import model_base_class

from typing import Optional, Union


class abundance_model(model_base_class):
    def __init__(self,
                 n_conditions: int,
                 n_subjects: int,
                 n_cell_types: int,
                 model_device: Optional[Union[str, torch.device]]=None):
        super().__init__()
        
        assert (n_cell_types > 1), "The number of cell types has to be > 1. Otherwise, there is nothing to estimate."
        assert (n_conditions > 1) or (n_subjects > 1), "You need to have at least 2 conditions or 2 subjects to estimate some effects."

        if model_device is None:
            self.model_device = torch.device(
                'cuda:0' if torch.cuda.is_available() else 'cpu')
        elif isinstance(model_device, str):
            self.model_device = torch.device(model_device)
        else:
            self.model_device = model_device

        self.n_conditions = n_conditions
        self.n_subjects = n_subjects
        self.n_cell_types = n_cell_types
        
        self.p_pi = p_effect_pi_class(model_device=self.model_device)
        self.p_pi_coef = p_effect_pi_coef_class(model_device=self.model_device,
                                                n_conditions=n_conditions,
                                                n_subjects=n_subjects,
                                                n_cell_types=n_cell_types)
        self.p_pi_hyper_coef = p_effect_pi_hyper_coef_class(model_device=self.model_device,
                                                            n_subjects=n_subjects,
                                                            n_cell_types=n_cell_types)
        self.q_pi_coef = q_effect_pi_coef_class(model_device=self.model_device,
                                                n_conditions=n_conditions,
                                                n_subjects=n_subjects,
                                                n_cell_types=n_cell_types)
        self.q_pi_hyper_coef = q_effect_pi_hyper_coef_class(model_device=self.model_device,
                                                            n_subjects=n_subjects,
                                                            n_cell_types=n_cell_types)
        
    def _update_distributions(self, 
                              one_hot_index: torch.tensor,
                              FC: torch.tensor,
                              RS: torch.tensor, 
                              mode="training"):
        with torch.set_grad_enabled(mode=="training"):
            # To generate all the distributions, we start from
            # the posteriors 
            # The steps are almost always: generate the distributions
            # Then sample from the distributions 
            ####################################
            # Q
            ################
            q_pi_coef_dict = self.q_pi_coef()
            pi_coef_samples = self.q_pi_coef.get_samples()
            
            q_pi_hyper_coef_dict = self.q_pi_hyper_coef()
            pi_hyper_coef_samples = self.q_pi_hyper_coef.get_samples()
                 
            ####################################
            # P
            ##############
            p_pi_hyper_coef_dict = self.p_pi_hyper_coef()
            p_pi_coef_dict = self.p_pi_coef(eta_sample=pi_hyper_coef_samples['eta'])
            
            index = torch.argmax(one_hot_index, dim=1)
            p_pi_dict = self.p_pi(FC=FC,
                                  RS=RS,
                                  beta_sample=pi_coef_samples['beta'],
                                  gamma_sample=pi_coef_samples['gamma'])
            log_likelihood = p_pi_dict['pi'].log_prob(index)
            
            return {
                "q_pi_hyper_coef_dict": q_pi_hyper_coef_dict,
                "q_pi_coef_dict": q_pi_coef_dict,
                "p_pi_hyper_coef_dict": p_pi_hyper_coef_dict,
                "p_pi_coef_dict": p_pi_coef_dict,
                "p_pi_dict": p_pi_dict,
                "log_likelihood": log_likelihood
            }
    
    def compute_loss(self, 
                     distribution_dict: dict,
                     show_details: bool=False):
        reconstruction_error = distribution_dict['log_likelihood'].sum()

        kl_hyper_coef = kl_divergence(distribution_dict['q_pi_hyper_coef_dict']['eta'],
                                      distribution_dict['p_pi_hyper_coef_dict']['eta']).sum()
        kl_coef = 0 
        for dist in distribution_dict['q_pi_coef_dict']:
            kl_coef += kl_divergence(distribution_dict['q_pi_coef_dict'][dist],
                                     distribution_dict['p_pi_coef_dict'][dist]).sum()  
        
        elbo = reconstruction_error - kl_coef - kl_hyper_coef
        
        if show_details:
            print("="*25)
            print("likelihood is {}, kl_coef is {}, kl_hyper_coef is {}".format(reconstruction_error,
                                                                                kl_coef,
                                                                                kl_hyper_coef))
            print("="*25)
        
        return -elbo
